/* OpenCA libpki package
* (c) 2000-2007 by Massimiliano Pala and OpenCA Group
* All Rights Reserved
*
* ===================================================================
* Released under OpenCA LICENSE
*/

#include <libpki/pki.h>

#define BUFF_MAX_SIZE	2048

/*! \brief Frees the memory assiociated with a PKI_HTTP data structure */

void PKI_HTTP_free ( PKI_HTTP *rv ) {
	if ( !rv ) return;

	if ( rv->location ) PKI_Free ( rv->location );
	if ( rv->type ) PKI_Free ( rv->type );

	if ( rv->body ) PKI_MEM_free ( rv->body );
	if ( rv->head ) PKI_MEM_free ( rv->head );

	PKI_Free ( rv );

	return;
}


/*! \brief Allocates the memory for a new PKI_HTTP data structure */

PKI_HTTP * PKI_HTTP_new ( void ) {

	PKI_HTTP *ret = NULL;

	if((ret = PKI_Malloc ( sizeof( PKI_HTTP ))) == NULL ) {
		return ( NULL );
	}

	/* Standard Header Codes */
	ret->code = 0;
	ret->type = NULL;
	ret->location = NULL;

	/* Data */
	ret->body = NULL;
	ret->head = NULL;

	return ( ret );
}

/*! \brief Returns a PKI_HTTP from the content of a char * */

char * PKI_HTTP_get_header_txt ( char *orig_data, char *header) {

	char *tk = NULL, *pnt = NULL;
	char *ret = NULL;

	char *data = NULL;
	int found = 0;

	if( !orig_data || !header || 
		(strlen(orig_data) == 0) || (strlen(header) == 0)) return NULL;

	if(( data = strdup( orig_data)) == NULL ) {
		PKI_log_debug("Memory Error");
		return NULL;
	}

	// PKI_log_debug("HEADERS DATA =>%s<=", orig_data );

	for ( tk = strtok_r ( data, "\r\n", &pnt ); tk;
			tk = strtok_r ( NULL, "\r\n", &pnt )) {

		if ( tk == NULL ) break;

		if ( strncmp_nocase( tk, header, strlen(header)) == 0 ) {
			// PKI_log_debug("Found Token => %s", tk );
			found = 1;
			break;
		}
	}

	if ( found == 0 ) {
		// PKI_log_debug ( "TOKEN::%s NOT Found!", header);
		PKI_Free ( data );
		return NULL;
	}

	if((pnt = strchr( tk, ':' )) == NULL ) {
		PKI_Free ( data );
		return NULL;
	}
	pnt++;

	while ( (pnt != NULL ) && (*pnt == ' ' ) ) {
			pnt++;
	};

	if( pnt ) {
		ret = strdup( pnt );
	}

	PKI_Free ( data );

	return ret;
}


/*! \brief Returns a PKI_HTTP from the content of a PKI_MEM */

char * PKI_HTTP_get_header ( PKI_HTTP *http, char * header ) {

	if( !http || !http->head || !header ) return NULL;

	return PKI_HTTP_get_header_txt ( (char *)http->head->data, header);
}

/*
PKI_HTTP *PKI_HTTP_parse_mem ( PKI_MEM *pki_mem ) {

        int head_lines = 0;
        char *linebuf = NULL;
	size_t linesize = 0;

	PKI_HTTP *ret = NULL;

	BIO *in = NULL;

	if( !pki_mem || !pki_mem->data || pki_mem->size < 1 ) {
		return (NULL);
	}

	if( (in = BIO_new_mem_buf( pki_mem->data, (int) pki_mem->size )) 
								== NULL ) {
		return ( NULL );
	}

	if(( ret = PKI_HTTP_new()) == NULL ) {
		return ( NULL );
	}

        head_lines = 0;
	if(( ret->head = PKI_MEM_new_null()) == NULL ) {
		return ( NULL );
	}

	linesize = 64 * 1024;
	if (( linebuf = PKI_Malloc ( linesize + 1)) == NULL ) {
		PKI_log_err ( "Memory allocation error");
		return (NULL);
	}

        do {
                int read_code = 0;
		size_t read_size = 0;
		char *eol = NULL;

                memset(linebuf, '\x0', linesize);
                if((read_code = BIO_gets(in, linebuf, linesize)) < 0) {
			break;
                }

		read_size += (int) strlen(linebuf);
		PKI_MEM_add ( ret->head, linebuf, strlen(linebuf));

		if(eol = strchr ( linebuf, '\r' )) {
			*eol = '\x0';
		};

		if (eol = strchr( linebuf, '\n' )) {
			*eol = '\x0';
		};

		if( head_lines == 0 ) {
			PKI_log_debug("LINE (%d) => %s", strlen(linebuf), linebuf );
			if( sscanf(linebuf,"HTTP/%f %d", &ret->version, &ret->code ) > 0 ) {
				PKI_log_debug ( "VER => %f --- CODE => %d", 
						ret->version, ret->code );

				if ( ret->code != 200 ) {
					PKI_log_debug("HTTP returned code (%d)", ret->code );
				}
			} else if ( strncmp_nocase(linebuf, PKI_HTTP_METHOD_GET_TXT, 
				sizeof(PKI_HTTP_METHOD_GET_TXT) ) == 0 ) {
				ret->method = PKI_HTTP_METHOD_GET;
				ret->url = URL_new ( &linebuf[4] );
			} else if ( strncmp_nocase( linebuf, PKI_HTTP_METHOD_POST_TXT, 
				sizeof(PKI_HTTP_METHOD_POST)) == 0 ) {
					ret->method = PKI_HTTP_METHOD_POST;
			} else {
				ret->method = PKI_HTTP_METHOD_UNKNOWN;
			};
		}

		if( strncmp_nocase("Location:", linebuf, strlen("Location:"))
									== 0 ){
			char *pnt = NULL;

			if((pnt = strchr( linebuf, ' ' )) == NULL ) {
				pnt = strchr ( linebuf, ':' ) + 1;
			}
			while ( (pnt != NULL ) && ( *pnt == ' ' ) ) {
				pnt++;
			};

			if( pnt ) {
				size_t len = 0;

				len = strlen( pnt );

				ret->location = PKI_Malloc ( len + 1 );
				memcpy( ret->location, pnt, len );
				ret->location[len] = '\x0';
			}
		}

		if( strncmp_nocase( "Content-type:", linebuf,
				strlen("Content-type:")) == 0 ) {
			char *pnt = NULL;
			if ((pnt = strchr( linebuf, ' ')) == NULL ) {
				pnt = strchr( linebuf, ':' ) + 1;
			}
			while( (pnt != NULL ) && ( *pnt == ' ')) {
				pnt++;
			}
			if ( pnt ) {
				size_t len = 0;
				len = strlen (pnt);
				ret->type = PKI_Malloc ( len + 1);
				memcpy( ret->type, pnt, len );
				ret->type[len] = '\x0';
			}
		}
		head_lines++;
        } while( strncmp(linebuf, "\r\n", strlen(linebuf)) > 0 );

	BIO_free( in );

	if( linebuf ) PKI_Free ( linebuf );

	PKI_log_debug( "PKI_MEM  SIZE => %d bytes", pki_mem->size );
	PKI_log_debug( "HEAD SIZE => %d bytes", ret->head->size );
	PKI_log_debug( "HEAD DATA => %s", ret->head->data );
	PKI_log_debug( "METHOD => %d", ret->method );

	if ( ret->url ) {
		PKI_log_debug( "PKI_URL => %s", ret->url->path );
	}

	if(( ret->body = PKI_MEM_new_data ( 
			(size_t) (pki_mem->size) - (size_t) ret->head->size, 
				&pki_mem->data[ret->head->size] )) == NULL ) {
		PKI_log_err ("ERROR::Memory");
	} else {
		PKI_log_debug( "RET BODY => %d bytes", ret->body->size );
	}

	return ( ret );
}
*/

/* Internal version, can handle both HTTP and HTTPS */

PKI_HTTP *__pki_http_get_sock_s ( int fd, PKI_SSL *ssl, 
					int timeout, ssize_t max_size ) {

	PKI_HTTP * ret = NULL;

	char buf[1024];
	char * data = NULL;
	char * eoh = NULL;

	int max = 0;
	int read = 0;
	int head_size = 0;
	int data_size = 0;
	int cont_size = 0;

	if ( max_size <= 2*sizeof(buf) ) {
		max = 64*1024 - sizeof(buf);
	} else {
		max = max_size - sizeof(buf);
	}

	if((data = PKI_Malloc ( max + sizeof(buf) )) == NULL ) {
		PKI_log_debug("Memory Error");
		return NULL;
	}

	if((ret = PKI_HTTP_new()) == NULL ) {
		PKI_log_debug("Memory Error");
		return NULL;
	}

	data_size = 0;
	for ( ; ; ) {
		int i = 0;
		int start = 0;

		/* Let's read some data */
		if( !ssl ) {
			read = PKI_NET_read( fd, buf, sizeof(buf), timeout);
		} else {
			read = PKI_SSL_read( ssl, buf, sizeof(buf));
		}

		/* Fast And Ugly Hack to Speedup Data Read for HTTP */
		if ( read > 0 ) {
			if ( data_size < 4 ) {
				start = 0;
			} else {
				start = data_size - 4;
			}

			/* Add the buff to the data */
			memcpy(&data[data_size], buf, read );

			/* Update the Size of read data */
			data_size += read;
		}

		if( !ret->head ) {
			if ( (eoh = strstr(&data[i], "\r\n\r\n")) != NULL ) {
				char * pnt = NULL;

				head_size = eoh - data + 4;
				ret->head = PKI_MEM_new_data( head_size + 1, 
							(unsigned char *) data);
				ret->head->data[head_size] = '\x0';

				if((pnt = PKI_HTTP_get_header(ret, 
						"Content-Length")) != NULL) {
					cont_size = atoi(pnt);
					PKI_Free ( pnt );
				}
			}
		}

		/* If we got headers and content-size */
		if ( (head_size > 0) && (cont_size > 0) ) {
			if ( cont_size <= data_size - head_size ) break;
		}

		/* Error, or end of data */
		if ( read <= 0 ) {
			// PKI_log_debug("FINISHED READING => %d + %d (%d)",
			// 	head_size, cont_size, data_size );

			cont_size = data_size - head_size;
			break;
		}

	}

	if(strncmp_nocase(data, PKI_HTTP_METHOD_GET_TXT, 
				strlen(PKI_HTTP_METHOD_GET_TXT)) == 0) {
		ret->method = PKI_HTTP_METHOD_GET;
		ret->url = URL_new ( &data[4] );
	} else if (strncmp_nocase(data, PKI_HTTP_METHOD_POST_TXT,
				strlen(PKI_HTTP_METHOD_POST_TXT)) == 0) {
		ret->method = PKI_HTTP_METHOD_POST;
	} else if (strncmp_nocase(data, "HTTP/", 5) == 0 ) {
		if(sscanf(data,"HTTP/%f %d", &ret->version, &ret->code ) < 1 ) {
			PKI_log_debug("ERROR Parsing HTTP MSG");
			PKI_HTTP_free ( ret );
			return NULL;
		}
	} else {
		PKI_log_debug("ERROR::Malformed HTTP message => %s", data);
		PKI_HTTP_free ( ret );
		return NULL;
	}

	if ( data_size - head_size <= 4 ) {
		PKI_log_debug( "WARNING: No content!");
	} else {
		ret->body = PKI_MEM_new_data ( cont_size, 
				(unsigned char *) &data[data_size-cont_size] );
		PKI_log_debug("HTTP DATA => size (%d->%d)", 
			data_size-cont_size, cont_size );
	}
 
	ret->location = PKI_HTTP_get_header ( ret, "Location" );
	ret->type = PKI_HTTP_get_header ( ret, "Content-Type" );

	/* Free the buffer memory */
	PKI_Free ( data );

	return ( ret );
}

/* Internal function to support data retrieval from http/https */

int __url_get_data_http_s_url ( URL *url, char *data, ssize_t data_size,
			char *content_type, int method, int timeout, 
			ssize_t max_size,PKI_MEM_STACK **sk, PKI_SSL *ssl) {

	int socket          = -1;

	ssize_t len          = 0;

	char *my_cont_type	= "application/unknown";

	PKI_HTTP *http_rv    = NULL;
	int	      rv    = -1;
	int 	     ret    = PKI_OK;

	size_t max_len = 0;

	char *tmp  = NULL;
    
	char *head_get ="GET %s HTTP/1.0\r\n"
			"Host: %s\r\n"
			"User-Agent: LibPKI\r\n"
		        "Connection: close\r\n";

	char *head_post =    "POST %s HTTP/1.0\r\n"
			"Host: %s\r\n"
			"User-Agent: LibPKI\r\n"
		        "Connection: close\r\n"
			"Content-type: %s\r\n"
			"Content-Length: %d\r\n";

	char *head = NULL;

	if ( timeout < 0 ) timeout = 0;

	// PKI_log_debug ("HTTP::Method %d::Connecting to => %s", 
	// 						method, url->addr );

	if ( url->proto == URI_PROTO_HTTP ) {
		if((socket = PKI_NET_open ( url, 0 )) < 0 ) {
			/* Can not connect to the server */
			PKI_log_err ("Failed to connect to %s:%d", 
						url->addr, url->port);
			return PKI_ERR;
		}
	} else if ( url->proto == URI_PROTO_HTTPS ) {
		PKI_log_debug("Connecting to SSL socket (%p)", ssl );

		if((!ssl) && ((ssl = PKI_SSL_new( NULL )) == NULL)) {
			PKI_log_err("Failed to create a new PKI_SSL");
			return PKI_ERR;
		}

		if( PKI_SSL_connect_url ( ssl, url, timeout ) == PKI_ERR ) {
			PKI_log_err("Failed to connect to %s", url->url_s );
			goto err;
		}
	}

	if ( method == PKI_HTTP_METHOD_GET ) {
		head = head_get;
		max_len = strlen(head) + strlen( url->path ) + 
						strlen( url->addr ) + 1;
		tmp = PKI_Malloc ( max_len );
		len = (ssize_t) snprintf(tmp, max_len, head, url->path, 
						url->addr );
	} else if ( method == PKI_HTTP_METHOD_POST ) {
		head = head_post;
		if ( content_type ) my_cont_type = content_type;
		max_len = strlen(head) + strlen( url->path ) + 
				strlen( url->addr ) + strlen(my_cont_type) + 101;
		tmp = PKI_Malloc ( max_len );
		len = (ssize_t) snprintf(tmp, max_len, head, url->path, url->addr, 
					my_cont_type, data_size );
	} else {
		PKI_log_err ( "Method (%d) not supported!", method );
		if ( !ssl ) {
			PKI_NET_close ( socket );
		} else {
			PKI_SSL_close ( ssl );
		}

		return PKI_ERR;
	}

	// PKI_log_debug ( "SENDING HEADERS::\n%s\n", tmp );

	if ( url->ssl == 0 ) {
		if((rv = PKI_NET_write ( socket, tmp, (size_t) len )) < 0 ) {
			/* Error in SSL_write */
			PKI_log_err("Can not write HTTP header to socket");
			goto err;
		}
	} else {
		if(( rv = PKI_SSL_write ( ssl, tmp, (ssize_t) len )) < 0 ) {
			PKI_log_err ("Can not write to PKI_SSL (%s)",
				ERR_error_string(ERR_get_error(), NULL));
			goto err;
		}
	}
	PKI_Free ( tmp );

	if ( url->usr && url->pwd ) {

		tmp = PKI_Malloc ( strlen(url->usr) + strlen(url->pwd) + 100);

		sprintf( tmp, "Authentication: user %s:%s\r\n\r\n", 
							url->usr, url->pwd);
	} else {
		tmp = PKI_Malloc ( 5 );
		sprintf( tmp, "\r\n");
	}
	len = strlen( tmp );

	if ( url->ssl == 0 ) {
		if((rv=PKI_NET_write(socket, tmp, (size_t) len)) < 0) {
                       	/* Error in SSL_write */
                       	PKI_log_err("Can not write HTTP header to socket");
                        	goto err;
		}
	} else {
		if(( rv = PKI_SSL_write(ssl,tmp,(ssize_t)len )) < 0 ) {
			PKI_log_err ("Can not write to PKI_SSL (%s)",
					ERR_error_string(ERR_get_error(),NULL));
			goto err;
		}
	}

	if( method == PKI_HTTP_METHOD_POST ) {
		/* Send the data together with the header */
		if ( url->ssl == 0 ) {
			if((rv = PKI_NET_write(socket,data,data_size)) < 0) {
				PKI_log_err ("Can not write POST to socket.");
				goto err;
			}
		} else {
			if((rv = PKI_SSL_write(ssl,data,data_size)) < 0) {
				PKI_log_err ("Can not write POST to PKI_SSL");
				goto err;
			}
		}
	}
	
	if ( url->ssl == 0 ) {
		if((http_rv=PKI_HTTP_get_sock(socket,timeout,max_size))== NULL){
			PKI_log_err ("HTTP retrieval error");
			goto err;
		}
	} else {
		if((http_rv=PKI_HTTPS_get_sock(ssl,timeout,max_size))== NULL){
			PKI_log_err ("HTTPS retrieval error");
			goto err;
		}
	}

	// PKI_log_debug("HEADERS => %s", http_rv->head );
	// PKI_log_debug("BODY => %s", http_rv->body );

	PKI_log_debug("HTTP Return Code ==> %d", http_rv->code );

	if ( http_rv->code >= 400 ) {
		goto err;
	} else if ( http_rv->code >= 300 ) {
		/* Redirection - let's try that */

		if( !http_rv->location ) {
			PKI_log_debug ( "HTTP Redirection but no location "
						"provided!");
			goto err;
		}

		PKI_log_debug("HTTP Redirection Location ==> %s", 
							http_rv->location );

		if( strstr(http_rv->location, "://") != NULL ) {
			URL *url_tmp = NULL;

			if( strncmp_nocase( http_rv->location, url->url_s, 
					(int) strlen(http_rv->location)) == 0) {
				PKI_log_debug( "HTTP cyclic redirection!");
				goto err;
			}

			if(( url_tmp = URL_new ( http_rv->location )) == NULL) {
				PKI_log_debug("HTTP location is not a valid "
					"URI (%s)", http_rv->location );
				goto err;
			}

			ret = __url_get_data_http_s_url ( url_tmp, data, 
				data_size, content_type, method, timeout, 
							max_size, sk, ssl );

			if ( url_tmp ) URL_free ( url_tmp );
	
			goto end;

		} else {
			const char *prot_s = NULL;
			char new_url[2048];
			URL *my_new_url = NULL;

			prot_s = URL_proto_to_string ( url->proto );
			if( !prot_s ) goto err;

			snprintf(new_url, sizeof(new_url),"%s://%s%s",
				prot_s, url->addr, http_rv->location );

			PKI_log_debug("OLD URL [%s] --> NEW URL [%s]",
				url->url_s, new_url );

			if( strncmp_nocase( new_url, url->url_s, 
					(int) strlen ( new_url )) == 0 ) {
				PKI_log_debug( "HTTP cyclic redirection!");
				goto err;
			}

			// if ( http_rv ) {
			// 	if ( http_rv->location ) 
			// 			PKI_Free ( http_rv->location );
			// 	PKI_HTTP_free ( http_rv );
			// }

			// if( obj ) PKI_MEM_free ( obj );

			my_new_url = URL_new ( new_url );

			ret = __url_get_data_http_s_url ( my_new_url, data, 
					data_size, content_type, method, 
						timeout, max_size, sk, ssl );

			goto end;
		}
	} else if ( http_rv->code != 200 ) {
		PKI_log_debug( "HTTP Return code not manageable (%d)",
							http_rv->code );
		goto err;
	}

	/* If a Pointer was provided, we want the data back :D */
	if ( sk ) {
		if ( ! *sk ) {
			if((*sk = PKI_STACK_MEM_new()) == NULL ) {
				PKI_log_debug ("Memory Failure");
				goto err;
			}
		}

		/* Add the returned value to the stack */
		PKI_STACK_MEM_push ( *sk, PKI_MEM_dup(http_rv->body));
	}

end:
	if ( http_rv ) {
		PKI_HTTP_free ( http_rv );
	}

	return ret;

err:
	if ( tmp ) PKI_Free ( tmp );

	if ( ssl ) PKI_SSL_free ( ssl );

	if ( http_rv ) {
		PKI_HTTP_free ( http_rv );
	}

	return PKI_ERR;
}

int __url_get_data_http_s ( char * url_s, char *data, ssize_t data_size,
			char *cont_type, int method, int timeout,
			ssize_t max_size, PKI_MEM_STACK **sk, PKI_SSL *ssl ) {

	int ret = PKI_OK;
	URL *url = NULL;

	if ( !url_s ) return PKI_ERR;

	if((url = URL_new ( url_s )) == NULL ) {
		return PKI_ERR;
	}

	ret = __url_get_data_http_s_url ( url, data, data_size, cont_type,
			method, timeout, max_size, sk, ssl );

	URL_free ( url );

	return ret;

}

/*! \brief Returns a parsed PKI_HTTP from a socket (read) */

PKI_HTTP *PKI_HTTP_get_sock ( int fd, int timeout, ssize_t max_size ) {
	return __pki_http_get_sock_s ( fd, NULL, timeout, max_size );
}

/*! \brief Returns data from a socket (see PKI_NET_get_data() for details) */

/*
PKI_MEM *http_get_data_sock ( int fd, int timeout, ssize_t max_size ) {
	return PKI_NET_get_data ( fd, timeout, max_size );
}
*/

/*! \brief Sends a HTTP message to a URL and retrieve the response
 *
 * Sends (POST/GET) data to a url and (if a pointer to a mem stack
 * is provided) returns the received response. PKI_ERR is returned
 * in case of error, otherwise PKI_OK is returned.
 */

int URL_get_data_http_s_url ( URL *url, int timeout, ssize_t max_size,
					PKI_MEM_STACK **sk, PKI_SSL *ssl ) {

	return __url_get_data_http_s_url ( url, NULL, 0, NULL, 
			PKI_HTTP_METHOD_GET, timeout, max_size, sk, ssl );
}

/*! \brief Returns the data from an HTTP source by using the GET command */

int URL_get_data_http_s ( char *url_s, int timeout, ssize_t max_size,
					PKI_MEM_STACK **ret, PKI_SSL *ssl ) {

	URL *url = NULL;
	int rv = PKI_OK;

	if( !url_s ) return PKI_ERR;

	if((url = URL_new( url_s )) == NULL ) {
		return PKI_ERR;
	}

	if( url->proto != URI_PROTO_HTTP ) {
		if( url ) URL_free ( url );
		return PKI_ERR;
	}

	rv = URL_get_data_http_s_url ( url, timeout, max_size, ret, ssl );

	return rv;
}

/*! \brief Use POST method to transfer data via HTTP. If a pointer to a
 *         PKI_MEM_STACK (eg., &stack) is provided, the returned data isi
 *         added to it
 */

int URL_post_data_http_s ( char *url_s, char *data, ssize_t size, 
			char *content_type, int timeout, ssize_t max_size,
				PKI_MEM_STACK **ret_sk, PKI_SSL *ssl ) {

	URL *url = NULL;
	int ret = PKI_OK;

	if( !url_s || !data || !content_type ) {
		/* ERROR: All data are strictly required! */
		return PKI_ERR;
	}

	if((url = URL_new(url_s)) == NULL ) {
		/* Error in creating the URL structure */
		return PKI_ERR;
	}

	ret = URL_post_data_http_s_url ( url, data, size, content_type,
				timeout, max_size, ret_sk, ssl );

	if ( url ) URL_free (url);

	return ret;
}

/*! \brief Posts data to an HTTP URL and returns the response data */

int URL_post_data_http_s_url ( URL *url, char *data, ssize_t size, 
			char *content_type, int timeout, ssize_t max_size,
				PKI_MEM_STACK **ret_sk, PKI_SSL *ssl ) {

	if ( !url || !data || size <= 0 ) return PKI_ERR;

	return __url_get_data_http_s_url ( url, data, size,
                        content_type, PKI_HTTP_METHOD_POST, timeout, 
				max_size, ret_sk, ssl );

}

/*! \brief Internal function to get data from a PKI_SSL */

/*
PKI_MEM *https_get_data_sock ( PKI_SSL *ssl, ssize_t max_size ) {

        PKI_MEM *buf = NULL;

	char tmp_buff[BUFF_MAX_SIZE];

        ssize_t newsize  = 0;

        if( !ssl ) {
		PKI_log_err("Attempted to retrieve data from "
				"null SSL socket" );
		return NULL;
	}

        // get subject name from bio using recommended OpenSSL template
        if((buf = PKI_MEM_new_null()) == NULL ) {
		PKI_log_err( "%s:%d::ERROR::Can not allocate memory!", 
				__FILE__, __LINE__ );
		return ( NULL );
	}

	while( (newsize = PKI_SSL_read( ssl, tmp_buff, BUFF_MAX_SIZE )) != 0 ) {

		if( newsize < 0 ) {
			PKI_log_debug("%s:%d::ERROR::%s\n", __FILE__,
						__LINE__, strerror(errno));
			break;
		}

		if( (max_size > 0) && 
			((ssize_t) (buf->size) + newsize > max_size) ) {
			newsize = max_size - (ssize_t) buf->size;
			PKI_MEM_add( buf, tmp_buff, (size_t) newsize);
			break;
		};

		PKI_MEM_add ( buf, tmp_buff, (size_t) newsize );

	};

	if( buf->size <= 0 ) {
		PKI_log_debug("%s:%d::WARNING::No HTTPS data retrieved.",
							__FILE__, __LINE__ );

		if( buf ) PKI_MEM_free ( buf );
		buf = NULL;
	} else {
        	// Make sure the buf ends with a \x0 value
		// PKI_MEM_add( buf, "\x0", 1);
	}

	return ( buf );
}
*/

/*! \brief Returns a parsed PKI_HTTP from a PKI_SSL socket (read) */

PKI_HTTP *PKI_HTTPS_get_sock ( PKI_SSL *ssl, int timeout, ssize_t max_size ) {
	return __pki_http_get_sock_s ( -1, ssl, timeout, max_size );
}

/*
PKI_MEM_STACK *URL_get_data_https_url ( URL *url, ssize_t max_size ) {

	SSL	*ssl        = NULL;

	int len = 0;

	PKI_MEM *obj        = NULL;
	PKI_MEM_STACK *ret  = NULL;

	PKI_HTTP *http_rv    = NULL;
	int	      rv    = -1;

	size_t max_len = 0;
        char *tmp  = NULL;

        char *head =    "GET %s HTTP/1.0\r\n"
                        "Host: %s\r\n"
                        "User-Agent: LibPKI\r\n"
                        "Connection: close\r\n\r\n"
                        "\r\n";

	if(( ssl = PKI_SSL_new ( url )) == NULL ) {
		return ( NULL );
	}

	max_len = strlen(head) + strlen( url->path ) + strlen( url->addr ) + 1;
        tmp = PKI_Malloc ( max_len );

        len = snprintf(tmp, max_len, head, url->path, url->addr );

	if((rv = SSL_write ( ssl, tmp, len )) < 0 ) {
		PKI_log_err("Can not write to SSL socket::%s", 
				ERR_error_string(ERR_get_error(), NULL )	);
		goto err;
	}

	if((obj = https_get_data_sock( ssl, max_size )) == NULL ) {
		PKI_log_debug( "No data returned from SSL connection");
		goto err;
	};

	if( (http_rv = PKI_HTTP_parse_mem ( obj )) == NULL ) {
		PKI_log_err("HTTP::Error!");
		goto err;
	}

	if ( http_rv->code >= 400 ) {
		goto err;
	} else if ( http_rv->code >= 300 ) {
		char new_url[2048];

		if( http_rv->location == NULL ) {
			PKI_log_debug ( "HTTP ERROR - Redirect but no new "
						" location provided!");
			goto err;
		}

		if( strstr(http_rv->location, "://") != NULL ) {

			if( strncmp_nocase( http_rv->location,
				    url->url_s, (int) strlen(http_rv->location)) ==0){
				PKI_log_debug( "HTTP cyclic redirection!");
				goto err;
			}

			snprintf( new_url, sizeof(new_url), "%s",
							http_rv->location );

			ret = URL_get_data( new_url, 0, max_size );

			if ( http_rv ) {
				if ( http_rv->location ) {
					PKI_Free ( http_rv->location );
				};
				PKI_HTTP_free ( http_rv );
			}

			if( obj ) PKI_MEM_free ( obj );

			return ( ret );
		} else {
			const char *prot_s = NULL;

			prot_s = URL_proto_to_string ( url->proto );
			if( !prot_s ) return ( NULL );

			snprintf(new_url, sizeof(new_url),"%s://%s%s",
				prot_s, url->addr, http_rv->location );

			if( strncmp_nocase( new_url,
					url->url_s, (int) strlen(new_url)) ==0){
				PKI_log_err( "HTTP cyclic redirection!");
				goto err;
			}

			if ( http_rv ) {
				PKI_HTTP_free ( http_rv );
				http_rv = NULL;
			}

			if( obj ) PKI_MEM_free ( obj );

			return ( URL_get_data ( new_url, 0, max_size ));
		}
	} else if ( http_rv->code != 200 ) {
		PKI_log_debug( "HTTP code not understood [%d]",
							http_rv->code );
		goto err;
	}

	if((ret = PKI_STACK_MEM_new()) != NULL ) {
		// PKI_MEM *mm = NULL;

		// if((mm = PKI_MEM_new_null()) == NULL ) {
		// 	return ( NULL );
		// }

		// PKI_MEM_add( mm, (char *) &(obj->data[http_rv->size]), 
		// 		(size_t) obj->size - (size_t) http_rv->size );
		PKI_STACK_push( ret, PKI_MEM_dup( http_rv->body ) );
	}

	if( tmp ) PKI_Free ( tmp );

	if( obj ) PKI_MEM_free ( obj );

	return( ret );

err:
	if ( tmp ) PKI_Free ( tmp );
	if ( ssl ) PKI_SSL_free ( ssl );
	if ( obj ) PKI_MEM_free ( obj );

	if ( http_rv ) PKI_HTTP_free ( http_rv );
		
	return ( NULL );
}
*/

/*! \brief Sends a HTTP message to a URL and retrieve the response over PKI_SSL
 *
 * Sends (POST/GET) data to a url and (if a pointer to a mem stack
 * is provided) returns the received response. PKI_ERR is returned
 * in case of error, otherwise PKI_OK is returned.
 */

/*
int URL_get_data_http_s_url( URL *url, PKI_MEM_STACK **ret_sk, int timeout,
					ssize_t max_size, PKI_SSL *ssl )  {

	if( __url_get_data_http_s ( url, data, size, method,
		content_type, ret_sk, timeout, max_size, ssl ) == PKI_ERR){
		return NULL;
	}

	return sk;
}
*/

/*! \brief Returns the data from an HTTPS source by using the GET command */

/*
PKI_MEM_STACK *URL_get_data_http_s ( char *url_s, 
					ssize_t max_size, PKI_SSL *ssl ) {

	URL *url = NULL;
	PKI_MEM_STACK *ret = NULL;

	if( !url_s ) return (NULL);

	if((url = URL_new( url_s )) == NULL ) {
		return NULL;
	}

	if( url->proto != URI_PROTO_HTTPS ) {
		if( url ) URL_free ( url );
		return( NULL );
	}

	ret  = __url_get_data_http_s ( url, NULL, 0, PKI_HTTP_METHOD_GET,
			NULL, &ret, 0, max_size, NULL );

	if ( url ) URL_free ( url );

	return ret;
}
*/

/*! \brief Posts data to an HTTP/HTTPS url */

int URL_put_data_http_s ( char *url_s, char *data, ssize_t size, 
			char *content_type, int timeout, ssize_t max_size,
				PKI_MEM_STACK **ret_sk, PKI_SSL *ssl ) {

	URL *url = NULL;
	int ret = PKI_ERR;

	if( !url_s || !data || !content_type ) {
		// ERROR: All data are strictly required!
		return(PKI_ERR);
	}

	if((url = URL_new(url_s)) == NULL ) {
		// Error in creating the URL structure
		return(PKI_ERR);
	}
 
	ret = __url_get_data_http_s_url ( url, data, size, content_type, 
			PKI_HTTP_METHOD_POST, timeout, max_size, ret_sk, ssl );

	// ret = URL_post_data_https_url ( url, data, size, content_type, 
		// 		ret_sk, timeout, max_size, ssl );

	if ( url ) URL_free (url);

	return ( ret );
}
 
/*

int URL_post_data_https_url ( URL *url, char *data, ssize_t size, 
				char *content_type, PKI_MEM_STACK **ret_sk ) {

	// int sock;
	PKI_SSL *ssl = NULL;

	char *headers = "POST %s HTTP/1.0\r\nContent-Type: %s\r\n"
			"Content-Length: %d\r\nHost: %s\r\n"
			"Connection: close\r\n\r\n";

	char buffer[BUFF_MAX_SIZE];
	int rv;

	PKI_MEM *buf = NULL;
	PKI_MEM *mm = NULL;

	PKI_HTTP *http_rv = NULL;

	if( !url || !data || !content_type ) {
		// ERROR: All data are strictly required!
		return ( PKI_ERR );
	}

	snprintf( buffer, sizeof(buffer), headers, url->path, 
					content_type, size, url->addr );

	if((ssl = PKI_SSL_new (NULL)) == NULL ) {
		// Can not connect to the server
		PKI_log_debug("Can not create SSL socket (%s:%d)",
					__FILE__, __LINE__ );
		return ( PKI_ERR );
	}

	if( PKI_SSL_connect( ssl, url->url_s, 0 ) == PKI_ERR ) {
		PKI_log_err("Can not connect to %s", url->url_s );
		PKI_SSL_free ( ssl );
		return PKI_ERR;
	}

	if((rv = PKI_SSL_write( ssl, buffer, (int) strlen( buffer ))) < 1 ) {
		PKI_log_err( "Can not write to SSL socket");
		PKI_SSL_free ( ssl );
		return ( PKI_ERR );
	}

	if((rv = PKI_SSL_write( ssl, data, size )) < 1 ) {
		// Error in writing the data
		PKI_log_err( "Can not write to SSL socket");
		PKI_SSL_free ( ssl );
		return ( PKI_ERR );
	}

	buf = https_get_data_sock( ssl, size );
	PKI_log_debug( "RECEIVED buf = %s", buf->data );

	// Free the SSL connection and data structures
	PKI_SSL_free ( ssl );

	if( (http_rv = PKI_HTTP_parse_mem ( buf )) == NULL ) {
		PKI_log_err("Can not parse HTTPS headers!");
		if( buf ) PKI_MEM_free ( buf );
		return ( PKI_ERR );
	}

	if( http_rv->code != 200 ) {
		PKI_log_debug("HTTP::Object not found (%d)", http_rv->code );
	}

	if( ret_sk ) {
		if( *ret_sk ) {
			PKI_STACK_MEM_free_all( *ret_sk );
		}

		if(( *ret_sk = PKI_STACK_MEM_new()) == NULL ) {
			return ( PKI_ERR );
		}

		PKI_STACK_MEM_push( *ret_sk, PKI_MEM_dup(http_rv->body));
	}

	return( PKI_OK );
}
*/
